/* This is the bootsector of the system. It uses the BIOS to load the kernel
 * image from a FAT12-formatted, 1.44MB floppy disk.
 */

	.code16
	.text
	.globl _start
	.org 0

#define error( x )	movw $x, %si; call fatal;

_start:
	jmp main
	nop

/* The BIOS parameter block - it is part of the fAT12 filesystem
 * Note that we are using some of its fields as variables to gain space
 */
	oem_name:				.ascii	"ABSTRACT"
	bytes_per_sector:		.word	512
	sectors_per_cluster:	.byte	1
	reserved_sectors:		.word	1
	num_fats:				.byte	2
	root_entries:			.word	224
	total_sectors_16:		.word	2880
	media:					.byte	0xF0
	fat_size:				.word	9
	sectors_per_track:		.word	18
	heads_per_cylinder:		.word	2
	hidden_sectors:			.long	0
	/* NOTE: The next two fields are normally a four-byte field which is the
	 * 32-bit number of sectors. Since we use the 16-bit one, we must put this
	 * field to 0. Here, we are using it as 2 word-sized variables to gain space
	 */
	first_data_sector:		.word	0x0000
	cluster:				.word	0x0000
	bootdrv:				.byte	0x00
	reserved:				.byte	0
	boot_signature:			.byte	0x29
	volume_id:				.long	0x1E1913C0
	volume_label:			.ascii	"SAMPLEGFXOS"
	filesystem_type:		.ascii	"FAT12   "


/* Here starts the real code */
main:
	/* First, we setup the segments and stack. The code and data sgments are set
	 * to zero and the offsets are used directly. We disable interrupts during
	 * that step, because a single interrupt while the stack is invalid will
	 * crash the system before it is even started.
	 */
	cli						/* Disable interrupts */
	xorw %ax, %ax			/* Equivalent of movw $0, %ax */
	movw %ax, %ds			/* Use this value for all segments */
	movw %ax, %es
	movw %ax, %ss
	movw $0x7C00, %sp		/* The stack is just below this bootsector */
	
	/* Some BIOS will load us at 0000:7C00 and others will load us at 07C0:0000.
	 * While both are the same address, it might confuse some programs. To avoid
	 * such trouble, we setup the code segment explicitly using this far jump.
	 */
	ljmp $0x0000, $1f
1:	movb %dl, bootdrv		/* Save the boot drive number */
	sti						/* Interrupts are safe again */
	

/* Now, we need to know the geometry of the disk. That is, its number of sectors
 * per track, tracks per cylinder, and number of heads. While these values are
 * already in the BPB, I prefer to ask them to the BIOS, because the BPB may
 * contain wrong data, while the BIOS gets it from the hardware, so it's sure.
 */
detect_geometry:
	movb $8, %ah			/* See http://en.wikipedia.org/wiki/INT_13H */
	pushw %es				/* Some BIOS' ints alter ES, as Bochs 2.6 */
	int $0x13
	popw %es
	jc fatal				/* This is in case of an error */
	andw $63, %cx
	shrw $8, %dx
	incw %dx
	movw %cx, sectors_per_track
	movw %dx, heads_per_cylinder

/* Okay, so let's start with the serious things. Here, we will load the root
 * directory of the filesystem. In that directory, we will look for out kernel
 * image, and load it just after that bootsector, at address 0x7E00.
 */
load_root_dir:
	/* Compute the size of the root directory. Normally, the formula is
	 * (32 * EntryCount)/BytesPerSector, but since floppies only have 512 bytes
	 * sectors, this simplifies to EntryCount/16 or EntryCount >> 4
	 */
	movw root_entries, %cx
	shrw $4, %cx
	
	/* Compute the first sector of the root directory. Normaly, the formula is
	 * (FatNum * FatSize) + Reserved, but most media have exactly 2 FATs, we can
	 * simplify this to: (FatSize << 1) + Reserved.
	 */
	movw fat_size, %ax
	shlw $1, %ax
	addw reserved_sectors, %ax
	
	/* Adding the root start to its size brings the first data sector, which we
	 * will need later. Store it, to avoid recalculating it then
	 */
	movw %ax, first_data_sector
	addw %cx, first_data_sector
	
	/* Load the root directory right after this program */
	movw $0x7E00, %bx
	call read_sectors
	
/* Now the root directory is in memory, we can start searching for it. Every
 * entry is 32 bytes, in which the first 11 bytes are the filename. We compare
 * it with our filename, until we find one.
 */
search_root_dir:
	movw root_entries, %cx		/* Repeat the search for each entry */
	movw $0x7E00, %di			/* The directory is loaded at 0x7E00 */
1:	pushw %cx					/* Save CX, it will be used inside the loop */
	pushw %di					/* As well as di, for the same reason */
	movw $11, %cx				/* Each filename is 8.3 -> 11 characters */
	movw $str_filename, %si		/* Compare those with the name of our file */
	rep cmpsb					/* Perform the comparison */
	popw %di					/* Pop back those before testing the results */
	popw %cx					/* of the comparison, for stack balance */
	je load_fat					/* Proceed if we found the file */
	addw $32, %di				/* Otherwise, skip to the next 32-bytes entry */
	loop 1b						/* And loop until there are no more entries */
	error( msg_nofile )			/* If we didn't find the file, bail out */

/* Now that we have our entry, we need to load the file. To do that, we load the
 * FAT at 0x1000, and use its information to load the file at 0x8000.
 */
load_fat:
	/* Save the number of the first cluster of our file. It is located at offset
	 * 26 in our root directory entry.
	 */
	movw 26(%di), %dx
	movw %dx, cluster
	
	/* Load only the first FAT: the second is just a copy of it and we aren't
	 * even sure that it exists.
	 */
	movw reserved_sectors, %ax	/* The FAT starts after the reserved sectors */
	movw fat_size, %cx			/* Its size (in sectors) is given by the BPB */
	movw $0x1000, %bx			/* Load at 0x1000 (ES is still 0x0000) */
	call read_sectors			/* Perform the loading */
	
	pushw $0x0000				/* Start with offset 0 */
	pushw $0x0800				/* And use ES to address directly 0x8000 */
	popw %es
	
/* Now, we're going to load the image cluster by cluster. We follow the cluster-
 * chains using the FAT that we loaded at 0x1000
 */
load_image:
	/* Retrieve the cluster number */
	movw cluster, %ax
	
	/* Calculate the address of the first sector in that cluster */
	subw $2, %ax
	movzx sectors_per_cluster, %cx
	mulw %cx
	addw first_data_sector, %ax
	
	/* Setup CX and BX to load all sectors in that cluster */
	popw %bx
	movzx sectors_per_cluster, %cx
	call read_sectors
	
	/* read_sectors incremented BX for us, so we just have to push it back */
	pushw %bx

/* Right, now we need to find the next cluster */
find_next_cluster:
	/* Calculate its index in the FAT */
	movw cluster, %ax
	movw %ax, %cx
	movw %ax, %dx
	shrw $1, %dx
	addw %dx,%cx /* CX = (3/2) cluster */
	
	/* Fetch that number from the FAT */
	movw $0x1000, %bx			/* Our FAT is loaded at 0x1000 */
	addw %cx, %bx				/* Add the index for the memory address */
	movw (%bx), %dx				/* Load that number into AX */
	
	/* FAT12 cluster numbers are 12 bits (1.5 bytes) so we need to cut the word
	 * that we read from memory, or we will have part of another cluster number
	 */
	testw $1, %ax
	jnz 1f
	andw $0x0FFF, %dx	/* Even cluster number: keep the low part */
	jmp 2f
1:	shrw $4, %dx		/* Odd cluster number: keep the high part */
2:	movw %dx, cluster	/* Save the next cluster's number */
	cmpw $0x0FFF, %dx	/* 0xFFF means the end of the cluster-chain */
	jne load_image		/* So unless it is found, we go loading next cluster */

/* Okay, when we arrive there, then we're done loading the file. All we have to
 * do now is to enable protected (32-bit) mode, and give control to the program
 * that we have loaded.
 */
we_are_done:
	movw $0x0013, %ax
	int $0x10
	ADDR32 lgdt gdt		/* Load a minimal GDT */
	movl %cr0, %eax		/* Enable protected mode */
	orb $1, %al
	movl %eax, %cr0
	ljmp $0x08, $1f		/* Far jump to the 32bit code segment */
	.code32
1:	movw $0x10, %ax		/* Reset all segments */
	movw %ax, %ds
	movw %ax, %es
	movw %ax, %ss
	movzx %sp, %esp
	//movl $0x7C00, %esp
	ljmp $0x08, $0x8000	/* And jump to the loaded code ! */

/* The functions above are for use in 32bit, get back to 16-bit mode */
.code16

/* This is an utility function. It converts an LBA addressector address (LBA is
 * a way to address sectors where the first sector is 0, the second 1 and so on)
 * to CHS addresses (given in terms of Sectors/Heads/Tracks).
 */
convert_lba:
	xorw %dx, %dx
	divw sectors_per_track
	incw %dx
	pushw %dx
	xorw %dx, %dx
	divw heads_per_cylinder
	movw %ax, %cx
	andw $0xFF, %cx
	shlw $8, %cx
	andw $768, %ax
	shrw $2, %ax
	orw %ax, %cx
	popw %ax
	orw %ax, %cx
	movb %dl, %dh
	ret

/* This function loads sectors from the boot disk. It reads %cx sectors starting
 * at sector %ax (given in LBA), to %es:%bx
 */
read_sectors:
1:	movw $5, %di		/* Retry each sector 5 times, floppies often fail */
2:	pushw %ax			/* Save those registers from BIOs modifications */
	pushw %bx
	pushw %cx
	pushw %dx
	pushw %es
	call convert_lba	/* Convert this to a CHS address */
	movw $0x0201, %ax	/* AH = 0x02 AL = 0x01 */
	movb bootdrv, %dl	/* Load from boot drive */
	int $0x13
	jnc 3f				/* In case of success, skip the following section */
	xorw %ax, %ax
	int $0x13			/* In case of failure, reset the disk */
	decw %di			/* Count this one try */
	popw %es			/* Retrieve all register values... */
	popw %dx
	popw %cx
	popw %bx
	popw %ax
	jnz 2b				/* And try again while tries != 0 */
	error( msg_disk )	/* If one sector failed to be read 5 times, bail out */
3:	popw %es			/* In case of success, retrieve all register values */
	popw %dx
	popw %cx
	popw %bx
	popw %ax
	addw bytes_per_sector, %bx	/* Increment the destination pointer */
	incw %ax			/* Increment AX so we load the next sector next time */
	loop 1b				/* And loop until sectors remain to be loaded */
	ret

/* This utility function is called when an error occurs. It prints and error
 * message, waits for the user to press a key, and reboots the computer.
 */
1:	movb $0x0E, %al /* BIOS function 0x0E of int 0x10: print character */
	int $0x10		/* This effectively prints the character */
fatal:
	lodsb				/* Load the byte at DS:SI into AL (our character) */
	testb %al, %al		/* Was this the null terminator ? */
	jnz 1b				/* If it wasn't, go printing the next character */
	xorw %ax, %ax
	int $0x16
	int $0x19
	cli
	hlt

/* Here are stored the constants used in the program */
variables:
	msg_nofile:			.asciz "No file"
	msg_disk:			.asciz "Disk fail"
	str_filename:		.ascii "SAMPLEOSIMG" /* This means SAMPLEOS.IMG */

/* This is the GDT we are going to use under protected mode. Because space is
 * a precious resource here, we are using the null descriptor (which the CPU
 * never looks at anyways) as the GDT pointer.
 */
gdt:
	/* [0x00] Null descriptor (used as GDT pointer) */
	.word gdtend - gdt - 1
	.long gdt
	.word 0
	/* [0x08] 32-bit code segment */
	.word 0xFFFF, 0x0000
	.byte 0x00, 0x9A, 0xCF, 0x00
	/* [0x10] 32-bit data/stack segment */
	.word 0xFFFF, 0x0000
	.byte 0x00, 0x92, 0xCF, 0x00
gdtend:

/* Put the boot signature in the last 2 bytes */
	. = _start + 510
	.word 0xAA55